package limax.provider;

import java.util.ArrayList;
import java.util.Collection;

import limax.codec.CodecException;
import limax.net.Protocol;
import limax.net.ServerManager;
import limax.net.SizePolicyException;
import limax.net.Transport;
import limax.util.Trace;
import limax.zdb.Procedure;
import limax.zdb.Transaction;

public final class ProcedureHelper {

	private ProcedureHelper() {
	}

	public static interface IOnlines {
		public void response(Protocol THIS, Protocol p)
				throws InstantiationException, SizePolicyException, CodecException, ClassCastException;

		public void unicast(long sessionid, Protocol p)
				throws InstantiationException, SizePolicyException, CodecException, ClassCastException;

		public void multicast(Collection<Long> sids, Protocol p)
				throws InstantiationException, SizePolicyException, CodecException, ClassCastException;

		public void broadcast(Protocol p)
				throws InstantiationException, SizePolicyException, CodecException, ClassCastException;
	}

	private static volatile IOnlines onlines = null;

	public static void setOnlines(IOnlines onlines) {
		ProcedureHelper.onlines = onlines;
	}

	public static IOnlines createIOnlines(ServerManager manager) {
		if (manager instanceof ProviderManagerImpl)
			return ((ProviderManagerImpl) manager).createIOnlines();
		else
			return null;
	}

	private static class UnicastTask implements Runnable {
		private final long key;
		private final Protocol p;

		public UnicastTask(long key, Protocol p) {
			this.key = key;
			this.p = p;
		}

		@Override
		public void run() {
			try {
				ProcedureHelper.onlines.unicast(key, p);
			} catch (InstantiationException | SizePolicyException | CodecException e) {
				if (Trace.isErrorEnabled())
					Trace.error(this, e);
			}
		}
	}

	private static class MulticastTask implements Runnable {
		private final Collection<Long> keys;
		private final Protocol p;

		public MulticastTask(Collection<Long> keys, Protocol p) {
			this.keys = new ArrayList<>(keys);
			this.p = p;
		}

		@Override
		public void run() {
			try {
				ProcedureHelper.onlines.multicast(keys, p);
			} catch (InstantiationException | SizePolicyException | CodecException e) {
				if (Trace.isErrorEnabled())
					Trace.error(this, e);
			}
		}
	}

	private static class ResponseTask implements Runnable {
		private final Protocol pFromSwitcher;
		private final Protocol p;

		public ResponseTask(Protocol pFromSwitcher, Protocol p) {
			this.pFromSwitcher = pFromSwitcher;
			this.p = p;
		}

		@Override
		public void run() {
			try {
				ProcedureHelper.onlines.response(pFromSwitcher, p);
			} catch (InstantiationException | SizePolicyException | CodecException e) {
				if (Trace.isErrorEnabled())
					Trace.error(this, e);
			}
		}
	}

	private static class BroadcastTask implements Runnable {
		private final Protocol p;

		public BroadcastTask(Protocol p) {
			this.p = p;
		}

		@Override
		public void run() {
			try {
				ProcedureHelper.onlines.broadcast(p);
			} catch (InstantiationException | SizePolicyException | CodecException e) {
				if (Trace.isErrorEnabled())
					Trace.error(this, e);
			}
		}
	}

	private static class SendDirectly implements Runnable {
		private final Transport transport;
		private final Protocol p;

		public SendDirectly(Transport transport, Protocol p) {
			this.transport = transport;
			this.p = p;
		}

		@Override
		public void run() {
			try {
				p.send(transport);
			} catch (InstantiationException | SizePolicyException | CodecException e) {
				if (Trace.isErrorEnabled())
					Trace.error(this, e);
			}
		}
	}

	public static void sendWhileFinish(long key, Protocol p) {
		Runnable r = new UnicastTask(key, p);
		Transaction.addSavepointTask(r, r);
	}

	public static void sendWhileFinish(Collection<Long> keys, Protocol p) {
		Runnable r = new MulticastTask(keys, p);
		Transaction.addSavepointTask(r, r);
	}

	public static void sendResponseWhileFinish(Protocol pFromSwitcher, Protocol p) {
		Runnable r = new ResponseTask(pFromSwitcher, p);
		Transaction.addSavepointTask(r, r);
	}

	public static void broadcastWhileFinish(Protocol p) {
		Runnable r = new BroadcastTask(p);
		Transaction.addSavepointTask(r, r);
	}

	public static void sendWhileFinish(Transport transport, Protocol p) {
		Runnable r = new SendDirectly(transport, p);
		Transaction.addSavepointTask(r, r);
	}

	public static void executeWhileFinish(Runnable r) {
		Transaction.addSavepointTask(r, r);
	}

	public static void executeWhileFinish(Procedure p) {
		executeWhileFinish(() -> Procedure.execute(p));
	}

	public static void sendWhileCommit(long key, Protocol p) {
		Transaction.addSavepointTask(new UnicastTask(key, p), null);
	}

	public static void sendWhileCommit(Collection<Long> keys, Protocol p) {
		Transaction.addSavepointTask(new MulticastTask(keys, p), null);
	}

	public static void sendResponseWhileCommit(Protocol pFromSwitcher, Protocol p) {
		Transaction.addSavepointTask(new ResponseTask(pFromSwitcher, p), null);
	}

	public static void broadcastWhileCommit(Protocol p) {
		Transaction.addSavepointTask(new BroadcastTask(p), null);
	}

	public static void sendWhileCommit(Transport transport, Protocol p) {
		Transaction.addSavepointTask(new SendDirectly(transport, p), null);
	}

	public static void executeWhileCommit(Runnable r) {
		Transaction.addSavepointTask(r, null);
	}

	public static void executeWhileCommit(Procedure p) {
		executeWhileCommit(() -> Procedure.execute(p));
	}

	public static void sendWhileRollback(long key, Protocol p) {
		Transaction.addSavepointTask(null, new UnicastTask(key, p));
	}

	public static void sendWhileRollback(Collection<Long> keys, Protocol p) {
		Transaction.addSavepointTask(null, new MulticastTask(keys, p));
	}

	public static void sendResponseWhileRollback(Protocol pFromSwitcher, Protocol p) {
		Transaction.addSavepointTask(null, new ResponseTask(pFromSwitcher, p));
	}

	public static void broadcastWhileRollback(Protocol p) {
		Transaction.addSavepointTask(null, new BroadcastTask(p));
	}

	public static void sendWhileRollback(Transport transport, Protocol p) {
		Transaction.addSavepointTask(null, new SendDirectly(transport, p));
	}

	public static void executeWhileRollback(Runnable r) {
		Transaction.addSavepointTask(null, r);
	}

	public static void executeWhileRollback(Procedure p) {
		executeWhileRollback(() -> Procedure.execute(p));
	}

	public static Procedure nameProcedure(String name, Procedure p) {
		return new Procedure() {

			@Override
			public boolean process() throws Exception {
				return p.process();
			}

			@Override
			public String getName() {
				return name;
			}
		};
	}

}
