package limax.zdb;

import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import limax.codec.MarshalException;
import limax.codec.OctetsStream;

final class TStorage<K, V> implements StorageInterface {
	private StorageEngine engine;

	TStorage(LoggerEngine logger, String tableName) {
		switch (Zdb.meta().getEngineType()) {
		case MYSQL:
			engine = new StorageMysql((LoggerMysql) logger, tableName);
			break;
		case EDB:
			engine = new StorageEdb((LoggerEdb) logger, tableName);
			break;
		default:
			throw new XError("unknown engine type");
		}
	}

	@Override
	public StorageEngine getEngine() {
		return engine;
	}

	private ConcurrentMap<K, TRecord<K, V>> changed = new ConcurrentHashMap<>();
	private ConcurrentMap<K, TRecord<K, V>> marshal = new ConcurrentHashMap<>();
	private ConcurrentMap<K, TRecord<K, V>> snapshot = new ConcurrentHashMap<>();
	private final ReentrantReadWriteLock snapshotClearLock = new ReentrantReadWriteLock();

	// from TTable
	void onRecordChange(TRecord<K, V> r) {
		if (r.getState() == TRecord.State.REMOVE) {
			changed.remove(r.getKey());
			marshal.remove(r.getKey());
		} else {
			changed.put(r.getKey(), r);
		}
	}

	// from TTableCache.Cleaner
	boolean isChangedOrUnknown(K key) {
		return changed.containsKey(key) || marshal.containsKey(key);
	}

	private volatile long countMarshalN = 0;
	private volatile long countMarshal0 = 0;
	private volatile long countFlush = 0;
	private volatile long countSnapshot = 0;
	private volatile long countMarshalNTryFail = 0;
	volatile long flushKeySize = 0;
	volatile long flushValueSize = 0;

	@Override
	public int marshalN() {
		int marshaled = 0;
		int tryFail = 0;
		for (Iterator<TRecord<K, V>> it = changed.values().iterator(); it.hasNext();) {
			TRecord<K, V> r = it.next();
			if (r.tryMarshalN()) {
				marshal.put(r.getKey(), r);
				it.remove();
				++marshaled;
			} else
				++tryFail;
		}
		this.countMarshalN += marshaled;
		this.countMarshalNTryFail += tryFail;
		return marshaled;
	}

	@Override
	public int marshal0() {
		marshal.putAll(changed);
		for (TRecord<K, V> r : changed.values())
			r.marshal0();
		final int size = changed.size();
		changed.clear();
		this.countMarshal0 += size;
		return size;
	}

	@Override
	public int snapshot() {
		final ConcurrentMap<K, TRecord<K, V>> tmp = this.snapshot;
		this.snapshot = this.marshal;
		this.marshal = tmp;
		for (TRecord<K, V> r : this.snapshot.values())
			r.snapshot();
		this.countSnapshot += snapshot.size();
		return snapshot.size();
	}

	@Override
	public int flush() {
		int count = 0;
		for (TRecord<K, V> r : this.snapshot.values()) {
			if (r.flush(this))
				++count;
		}
		Map<K, TRecord<K, V>> tmp = null;
		final Lock writeLock = snapshotClearLock.writeLock();
		writeLock.lock();
		try {
			tmp = this.snapshot;
			this.snapshot = new ConcurrentHashMap<>();
		} finally {
			writeLock.unlock();
		}
		for (TRecord<K, V> r : tmp.values())
			r.clear();
		this.countFlush += count;
		return count;
	}

	final boolean exist(K key, TTable<K, V> table) {
		final Lock readLock = snapshotClearLock.readLock();
		readLock.lock();
		try {
			TRecord<K, V> r = this.snapshot.get(key);
			if (null != r)
				return r.exist();
		} finally {
			readLock.unlock();
		}
		return engine.exist(table.marshalKey(key));
	}

	V find(K key, TTable<K, V> table) {
		OctetsStream value = null;
		boolean foundInSnapshot = false;

		final Lock readLock = snapshotClearLock.readLock();
		readLock.lock();
		try {
			TRecord<K, V> r = this.snapshot.get(key);
			if (null != r) {
				foundInSnapshot = true;
				value = r.find();
			}
		} finally {
			readLock.unlock();
		}

		try {
			if (foundInSnapshot) {
				if (null != value)
					return table.unmarshalValue(value);
				return null;
			}
			value = engine.find(table.marshalKey(key));
			if (null != value)
				return table.unmarshalValue(value);
			return null;
		} catch (MarshalException me) {
			throw new RuntimeException(me);
		}
	}

	public long getCountFlush() {
		return countFlush;
	}

	public long getCountMarshal0() {
		return countMarshal0;
	}

	public long getCountMarshalN() {
		return countMarshalN;
	}

	public long getCountMarshalNTryFail() {
		return countMarshalNTryFail;
	}

	public long getCountSnapshot() {
		return countSnapshot;
	}

	public long getFlushKeySize() {
		return flushKeySize;
	}

	public long getFlushValueSize() {
		return flushValueSize;
	}
}
