#include "common.h"
#include "oshelper.h"
#include "luahelper.h"

class EndpointObject
{
	lua_State*	lua_state;
	int lua_ref_config = -1;
	int lua_ref_manager = -1;
	int lua_ref_variant_callback = -1;
	int lua_ref_variant = -1;
	int closeReason = 0;
	std::shared_ptr<limax::EndpointConfig> config;
	limax::EndpointManager* manager = nullptr;
	limax::LuaVariantContextPtr variant;
	volatile bool isLoginDone = false;

	class Listener : public limax::EndpointListener
	{
		EndpointObject* object;
	public:
		Listener(EndpointObject* _object)
			: object(_object)
		{}
	public:
		virtual void onAbort(limax::Transport*) override
		{
			auto l = object->lua_state;
			auto c = object->lua_ref_config;
			fireNoitfyNoParam(object->lua_state, object->lua_ref_config, "onAbort", [l, c](int s, int e, const std::string& m) { fireErrorOccured(l, c, s, e, m); });
		}
		virtual void onTransportRemoved(limax::Transport*) override
		{
			object->isLoginDone = false;

			auto l = object->lua_state;
			auto c = object->lua_ref_config;
			auto lsec = [l, c](int s, int e, const std::string& m) { fireErrorOccured(l, c, s, e, m); };
			fireNoitfyOneParam(object->lua_state, object->lua_ref_variant, "onclose", object->closeReason, lsec);
			fireNoitfyNoParam(object->lua_state, object->lua_ref_config, "onTransportRemoved", lsec);
		}
		virtual void onErrorOccured(int errorsrouce, int errorvalue, const std::string& info) override
		{
			if (limax::SOURCE_LIMAX == errorsrouce && info == "switcherendpoint::SessionKick")
				object->closeReason = errorvalue;
			fireErrorOccured(object->lua_state, object->lua_ref_config, errorsrouce, errorvalue, info);
		}

		virtual void onTransportAdded(limax::Transport*) override
		{
			object->isLoginDone = true;

			{
				lua_State* L = object->lua_state;
				GetGlobalObject _dummy_object(L, object->lua_ref_manager);

				lua_pushunsigned(L, object->manager->getSessionId());
				lua_setfield(L, -2, "sessionId");

				lua_pushinteger(L, object->manager->getAccountFlags());
				lua_setfield(L, -2, "accountFlags");

				const auto& pvids = object->config->getVariantProviderIds();
				if (!pvids.empty())
				{
					object->variant = limax::LuaCreator::createLuaVariantContext(L, object->lua_ref_config);
					object->variant->createEndpointManagerTable(object->manager, object->config);
					lua_pushvalue(L, -1);
					object->lua_ref_variant = appendObjectToGlobal(L);
					lua_setfield(L, -2, "variant");

					if (-1 != object->lua_ref_variant_callback)
					{
						getObjectFromGlobal(L, object->lua_ref_variant_callback);
						getObjectFromGlobal(L, object->lua_ref_variant);
						int ec = lua_pcall(L, 1, 0, 0);
						if (LUA_OK != ec)
						{
							const char* msg = lua_tostring(L, -1);
							onErrorOccured(limax::SOURCE_LUA, limax::LUA_CALL_LUA_FUNC_FAILED, msg);
							lua_pop(L, 1);
						}
						releaseObjectFromGlobal(L, object->lua_ref_variant_callback);
						object->lua_ref_variant_callback = -1;
					}
				}
			}
			auto l = object->lua_state;
			auto c = object->lua_ref_config;
			auto lsec = [l, c](int s, int e, const std::string& m) { fireErrorOccured(l, c, s, e, m); };
			fireNoitfyNoParam(object->lua_state, object->lua_ref_config, "onTransportAdded", lsec);
			fireNoitfyNoParam(object->lua_state, object->lua_ref_variant, "onopen", lsec);
		}

		virtual void onKeepAlived(int ping) override
		{
			auto l = object->lua_state;
			auto c = object->lua_ref_config;
			fireNoitfyOneParam(object->lua_state, object->lua_ref_config, "onKeepAlived", ping, [l, c](int s, int e, const std::string& m) { fireErrorOccured(l, c, s, e, m); });
		}

		virtual void onSocketConnected() override
		{
			auto l = object->lua_state;
			auto c = object->lua_ref_config;
			fireNoitfyNoParam(object->lua_state, object->lua_ref_config, "onSocketConnected", [l, c](int s, int e, const std::string& m) { fireErrorOccured(l, c, s, e, m); });
		}
		virtual void onKeyExchangeDone() override
		{
			auto l = object->lua_state;
			auto c = object->lua_ref_config;
			fireNoitfyNoParam(object->lua_state, object->lua_ref_config, "onKeyExchangeDone", [l, c](int s, int e, const std::string& m) { fireErrorOccured(l, c, s, e, m); });
		}

		virtual void onManagerInitialized(limax::EndpointManager* mng, limax::EndpointConfig*) override
		{
			object->manager = mng;
			auto l = object->lua_state;
			auto c = object->lua_ref_config;
			fireNoitfyNoParam(object->lua_state, object->lua_ref_config, "onManagerInitialized", [l, c](int s, int e, const std::string& m) { fireErrorOccured(l, c, s, e, m); });
		}

		virtual void onManagerUninitialized(limax::EndpointManager*) override
		{
			object->manager = nullptr;
			auto l = object->lua_state;
			auto c = object->lua_ref_config;
			fireNoitfyNoParam(object->lua_state, object->lua_ref_config, "onManagerUninitialized", [l, c](int s, int e, const std::string& m) { fireErrorOccured(l, c, s, e, m); });
		}

		virtual void destroy() override
		{
			delete this;
		}
	};
	friend class Listener;
	Listener* listener;

	static std::mutex	object_mutex;
	static std::unordered_set<EndpointObject*> object_set;
public:
	EndpointObject(lua_State* _lua_state)
		: lua_state(_lua_state), listener(nullptr)
	{}
	~EndpointObject()
	{
		lua_State* L = lua_state;
		releaseObjectFromGlobal(L, lua_ref_config);
		releaseObjectFromGlobal(L, lua_ref_manager);
		releaseObjectFromGlobal(L, lua_ref_variant);
		releaseObjectFromGlobal(L, lua_ref_variant_callback);
	}
public:
	void start(std::shared_ptr<limax::EndpointConfig> _config, int refconfig, int refmanager, int refvariant)
	{
		lua_ref_config = refconfig;
		lua_ref_manager = refmanager;
		lua_ref_variant_callback = refvariant;
		config = _config;
		listener = new Listener(this);
		limax::Endpoint::start(config, listener);

		std::lock_guard<std::mutex> __locker__(object_mutex);
		object_set.insert(this);
	}

	void release(lua_State* L)
	{
		if (manager)
		{
			manager->close();
			while (manager != nullptr)
			{
				limax::uiThreadSchedule();
				std::this_thread::sleep_for(std::chrono::milliseconds(1));
			}
		}
		{
			std::lock_guard<std::mutex> __locker__(object_mutex);
			object_set.erase(this);
		}
		delete this;
	}

	static void ReleaseObjects()
	{
		std::unordered_set<EndpointObject*> objs;
		{
			std::lock_guard<std::mutex> __locker__(object_mutex);
			object_set.swap(objs);
		}
		for (auto obj : objs)
			delete obj;
	}
};

std::mutex	EndpointObject::object_mutex;
std::unordered_set<EndpointObject*> EndpointObject::object_set;

int lua_context_object_close(lua_State* L)
{
	lua_getfield(L, -1, getLimaxUserDataName());
	EndpointObject* obj = (EndpointObject*)lua_touserdata(L, -1);
	lua_pop(L, 1);
	obj->release(L);
	return 0;
}

int lua_create_context_and_start_endpoint(lua_State* L)
{
	lua_pushstring(L, "ping");
	lua_gettable(L, -2);
	int isping = lua_toboolean(L, -1);
	lua_pop(L, 1);

	std::string serverip;
	int serverport;
	{
		lua_pushstring(L, "serverip");
		lua_gettable(L, -2);
		serverip = lua_check_string(L, -1);
		lua_pop(L, 1);

		lua_pushstring(L, "serverport");
		lua_gettable(L, -2);
		serverport = luaL_checkint(L, -1);
		lua_pop(L, 1);
	}

	auto object = std::unique_ptr<EndpointObject>(new EndpointObject(L));
	std::shared_ptr<limax::EndpointConfigBuilder> builder;
	if (isping)
	{
		builder = limax::Endpoint::createPingOnlyConfigBuilder(serverip, serverport);
	}
	else
	{
		lua_pushstring(L, "username");
		lua_gettable(L, -2);
		std::string username = lua_check_string(L, -1);
		lua_pop(L, 1);

		lua_pushstring(L, "token");
		lua_gettable(L, -2);
		std::string token = lua_check_string(L, -1);
		lua_pop(L, 1);

		lua_pushstring(L, "platflag");
		lua_gettable(L, -2);
		std::string paltflag = lua_check_string(L, -1);
		lua_pop(L, 1);

		builder = limax::Endpoint::createLoginConfigBuilder(serverip, serverport, username, token, paltflag);
	}

	int refvariantcallback = -1;
	{
		lua_pushstring(L, "variant");
		lua_gettable(L, -2);
		if (lua_istable(L, -1))
		{
			std::vector<int16_t> pvids;
			lua_pushfstring(L, "pvids");
			lua_gettable(L, -2);
			if (lua_istable(L, -1))
			{
				int n = (int)lua_rawlen(L, -1);
				pvids.reserve(n);
				for (int i = 1; i <= n; i++)
				{
					lua_rawgeti(L, -1, i);
					int pvid = luaL_checkint(L, -1);
					pvids.push_back((int16_t)pvid);
					lua_pop(L, 1);
				}
				builder->variantProviderIds(pvids);
			}
			lua_pop(L, 1);

			if (!pvids.empty())
			{
				lua_pushfstring(L, "callback");
				lua_gettable(L, -2);
				if (lua_isfunction(L, -1))
					refvariantcallback = appendObjectToGlobal(L);
				else
					lua_pop(L, 1);
			}
			else
			{
				luaL_error(L, "variant table need member : pvids(table) [, callback(function)]");
			}
		}
		lua_pop(L, 1);
	}

	lua_pushvalue(L, -1);
	int refconfig = appendObjectToGlobal(L);
	{
		lua_pushstring(L, "script");
		lua_gettable(L, -2);
		if (lua_istable(L, -1))
		{
			limax::ScriptEngineHandlePtr seh = limax::LuaCreator::createScriptEngineHandle(L, -1, [L, refconfig](int s, int e, const std::string& m) { fireErrorOccured(L, refconfig, s, e, m); });
			lua_pop(L, 1);
			if (seh)
				builder->scriptEngineHandle(seh);
		}
		lua_pop(L, 1);
	}
	builder->executor(limax::runOnUiThread);

	lua_newtable(L);
	lua_pushlightuserdata(L, object.get());
	lua_setfield(L, -2, getLimaxUserDataName());

	lua_pushcfunction(L, lua_context_object_close);
	lua_setfield(L, -2, "close");

	lua_pushvalue(L, -1);
	int refmanager = appendObjectToGlobal(L);

	object->start(builder->build(), refconfig, refmanager, refvariantcallback);
	object.release();
	return 1;
}

int lua_open_engine(lua_State* L)
{
	limax::Endpoint::openEngine();
	return 0;
}

int lua_close_engine(lua_State* L)
{
	volatile bool done = false;
	limax::Endpoint::closeEngine([&done](){ done = true; });
	while (!done)
	{
		limax::uiThreadSchedule();
		std::this_thread::sleep_for(std::chrono::milliseconds(1));
	}
	EndpointObject::ReleaseObjects();
	return 0;
}

int lua_idle_process(lua_State* L)
{
	limax::uiThreadSchedule();
	return 0;
}

extern "C" LIMAX_LUA_API int luaopen_limaxcontext(lua_State* L)
{
	lua_newtable(L);

	lua_pushcfunction(L, lua_open_engine);
	lua_setfield(L, -2, "openEngine");

	lua_pushcfunction(L, lua_close_engine);
	lua_setfield(L, -2, "closeEngine");

	lua_pushcfunction(L, lua_idle_process);
	lua_setfield(L, -2, "idle");

	lua_pushcfunction(L, lua_create_context_and_start_endpoint);
	lua_setfield(L, -2, "start");

	table_append_oshelper(L);
	return 1;
}

